/* $Header: ntp_io.c,v 1.3 91/06/19 11:07:15 ath Exp $ */
/*
 * xntp_io.c - input/output routines for xntpd running under NOS.
 */
#include <stdio.h>

#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "netuser.h"
#include "socket.h"
#include "sockaddr.h"
#include "udp.h"

#include "ntp_types.h"
#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"
#ifdef REFCLOCK
#include "ntp_refclock.h"
#endif /* REFCLOCK */

#if 0			   /* recvbuf memory management not needed for NOS */
/*
 * recvbuf memory management--modified for the PC & expected low traffic.
 */
#define	RECV_INIT	4	/* 10 buffers initially */
#define	RECV_LOWAT	2	/* when we're down to three buffers get more */
#define	RECV_INC	2	/* get 5 more at a time */
#define	RECV_TOOMANY	10	/* this is way too many buffers */

/*
 * Memory allocation
 */
u_long full_recvbufs;	/* number of recvbufs on fulllist */
u_long free_recvbufs;	/* number of recvbufs on freelist */

struct recvbuf *freelist;	/* free buffers */
struct recvbuf *fulllist;	/* buffers with data */

u_long total_recvbufs;	/* total recvbufs currently in use */
u_long lowater_additions;	/* number of times we have added memory */

struct recvbuf initial_bufs[RECV_INIT];	/* initial allocation */
#endif /* 0 */

/*
 * Other statistics of possible interest
 */
u_long packets_dropped;	/* total number of packets dropped on reception */
u_long packets_ignored;	/* packets received on wild card interface */
u_long packets_received;	/* total number of packets received */
u_long packets_sent;	/* total number of packets sent */
u_long packets_notsent;	/* total number of packets which couldn't be sent */

u_long handler_calls;	/* number of calls to interrupt handler */
u_long handler_pkts;	/* number of pkts received by handler */
u_long io_timereset;	/* time counters were reset */

/*
 * Interface stuff
 */
#define	MAXINTERFACES	10
struct interface *any_interface;	/* pointer to default interface */
struct interface *loopback_interface;	/* point to loopback interface */
struct interface inter_list[MAXINTERFACES];
int ninterfaces;

/*
 * Imported from ntp_timer.c
 */
/*extern u_long current_time;  Now a macro in ntp.h */

extern int debug;


/*
 * init_io - initialize I/O data structures and call socket creation routine
 */
void
init_io()
{
	register int i;

#if 0
	/*
	 * Init buffer free list and stat counters
	 */
	freelist = 0;
	for (i = 0; i < RECV_INIT; i++) {
		initial_bufs[i].next = freelist;
		freelist = &initial_bufs[i];
	}

	fulllist = 0;
	free_recvbufs = total_recvbufs = RECV_INIT;
	full_recvbufs = lowater_additions = 0;
#endif  /* 0 */

	packets_dropped = packets_received = 0;
	packets_ignored = 0;
	packets_sent = packets_notsent = 0;
	handler_calls = handler_pkts = 0;
	io_timereset = 0;
	loopback_interface = 0;


	/*
	 * Create the sockets
	 */
	(void) create_sockets(htons(NTP_PORT));

#ifdef DEBUG
	if (debug)
		printf("init_io: interfaces %d\n", ninterfaces);
#endif
}


/*
 * create_sockets - create a socket for each interface plus a default
 *		    socket for when we don't know where to send
 */
int
create_sockets(port)
	unsigned int port;
{
	struct udp_cb *cb;
	int n, i, j;
	struct sockaddr_in resmask;
	int open_socket();
	void input_handler();
	extern void restrict();

	/*
	 * create pseudo-interface with wildcard address
	 * Only used for sending packets, not receiving.
	 */
	inter_list[0].sin.sin_family = AF_INET;
	inter_list[0].sin.sin_port = port;
	inter_list[0].sin.sin_addr.s_addr = INADDR_ANY;
	inter_list[0].sock.port = htons(port);
	inter_list[0].sock.address = htonl(INADDR_ANY);

	(void) strncpy(inter_list[0].name, "wildcard",
		       sizeof(inter_list[0].name));
	inter_list[0].received = 0;
	inter_list[0].sent = 0;
	inter_list[0].notsent = 0;
/*	inter_list[0].flags = INT_BROADCAST;*/
	inter_list[0].cb = NULLUDP; /* Not needed because no receive. */

	{ extern struct iface *Ifaces;
	  extern struct iface Loopback;
	  struct iface *ifr;

	  i = 1;
	  for (ifr = Ifaces; ifr != NULLIF; ifr = ifr->next) {
		inter_list[i].flags = 0;
	  	/* I'm not sure that this is correct... */
	    	if (ifr->broadcast)
		  {
		    inter_list[i].flags |= INT_BROADCAST;
		    inter_list[i].bcast.sin_family = AF_INET;
		    inter_list[i].bcast.sin_port = port;
		    inter_list[i].bcast.sin_addr.s_addr =
		      htonl(ifr->broadcast);
		    inter_list[i].bsock.address = ifr->broadcast;
		    inter_list[i].bsock.port = ntohs(port);
		  }

		if (ifr == &Loopback)
		  {
		    inter_list[i].flags |= INT_LOOPBACK;
		    if (loopback_interface == 0)
		      loopback_interface = &inter_list[i];
		  }

		(void)strncpy(inter_list[i].name, ifr->name,
			      sizeof(inter_list[i].name));

		inter_list[i].sin.sin_family = AF_INET;
		inter_list[i].sin.sin_port = port;
		inter_list[i].sin.sin_addr.s_addr = htonl(ifr->addr);
		inter_list[i].sock.address = ifr->addr;
		inter_list[i].sock.port = ntohs(port);

		inter_list[i].mask.sin_addr.s_addr = htonl(ifr->netmask);

		/* 
		 * look for an already existing source interface address.  If
		 * the machine has multiple point to point interfaces, then 
		 * the local address may appear more than once.
		 */		   
		for (j=0; j < i; j++)
			if (inter_list[j].sock.address == 
			    inter_list[i].sock.address) {
				break;
			}
		if (j == i)
			i++;
	  }
	}
	ninterfaces = i;

	for (i = 0; i < ninterfaces; i++) {
		inter_list[i].cb = open_udp(&inter_list[i].sock,
					    input_handler);
		/* Put the index of this interface into the user field
		 * so that input_handler() can find the interface quickly.
		 */
		inter_list[i].cb->user = i;
	}

	/*
	 * Blacklist all bound interface addresses
	 */
	resmask.sin_addr.s_addr = ~0L;
	for (i = 1; i < ninterfaces; i++)
		restrict(RESTRICT_FLAGS, &inter_list[i].sin, &resmask,
		    RESM_NTPONLY|RESM_INTERFACE, RES_IGNORE);

	any_interface = &inter_list[0];
	return ninterfaces;
}


/*
 * io_setbclient - open the broadcast client sockets
 * For NOS, don't do anything but set the flag.
 */
void
io_setbclient()
{
	int i;

	for (i = 1; i < ninterfaces; i++) {
		if (!(inter_list[i].flags & INT_BROADCAST))
			continue;
		if (inter_list[i].flags & INT_BCASTOPEN)
			continue;
/*		inter_list[i].bfd = open_socket(&inter_list[i].bcast, 0);*/
		inter_list[i].flags |= INT_BCASTOPEN;
	}
}


/*
 * io_unsetbclient - close the broadcast client sockets
 * For NOS, don't do anything but clear the flag.
 */
void
io_unsetbclient()
{
	int i;

	for (i = 1; i < ninterfaces; i++) {
		if (!(inter_list[i].flags & INT_BCASTOPEN))
			continue;
/*		close_socket(inter_list[i].bfd);*/
		inter_list[i].flags &= ~INT_BCASTOPEN;
	}
}



/*
 * findbcastinter - find broadcast interface corresponding to address
 */
struct interface *
findbcastinter(addr)
	struct sockaddr_in *addr;
{
	register int i;
	register u_long netnum;

	netnum = NSRCADR(addr);
	for (i = 1; i < ninterfaces; i++) {
		if (!(inter_list[i].flags & INT_BROADCAST))
			continue;
		if (NSRCADR(&inter_list[i].bcast) == netnum)
			return &inter_list[i];
		if ((NSRCADR(&inter_list[i].sin) & NSRCADR(&inter_list[i].mask))
		    == (netnum & NSRCADR(&inter_list[i].mask)))
			return &inter_list[i];
	}
	return any_interface;
}


#if 0				/* Buffer management not needed for NOS */

/*
 * getrecvbufs - get receive buffers which have data in them
 *
 * ***N.B. must be called with SIGIO blocked***
 */
struct recvbuf *
getrecvbufs()
{
	struct recvbuf *rb;
	extern char *emalloc();

#ifdef DEBUG
	if (debug > 4)
		printf("getrecvbufs: %lu handler interrupts, %lu frames\n",
		    handler_calls, handler_pkts);
#endif

	if (full_recvbufs == 0) {
#ifdef DEBUG
		if (debug > 4)
			printf("getrecvbufs called, no action here\n");
#endif
		return (struct recvbuf *)0;	/* nothing has arrived */
	}
	
	/*
	 * Get the fulllist chain and mark it empty
	 */
#ifdef DEBUG
	if (debug > 4)
		printf("getrecvbufs returning %d buffers\n", full_recvbufs);
#endif
	rb = fulllist;
	fulllist = 0;
	full_recvbufs = 0;

	/*
	 * Check to see if we're below the low water mark.
	 */
	if (free_recvbufs <= RECV_LOWAT) {
		register struct recvbuf *buf;
		register int i;

		if (total_recvbufs >= RECV_TOOMANY)
			syslog(LOG_ERR, "too many recvbufs allocated (%d)",
			    total_recvbufs);
		else {
			buf = (struct recvbuf *)
			    emalloc(RECV_INC*sizeof(struct recvbuf));
			for (i = 0; i < RECV_INC; i++) {
				buf->next = freelist;
				freelist = buf;
				buf++;
			}

			free_recvbufs += RECV_INC;
			total_recvbufs += RECV_INC;
			lowater_additions++;
		}
	}

	/*
	 * Return the chain
	 */
	return rb;
}


/*
 * freerecvbuf - make a single recvbuf available for reuse
 */
void
freerecvbuf(rb)
	struct recvbuf *rb;
{
	int osig;

	BLOCKIO(osig);
	rb->next = freelist;
	freelist = rb;
	free_recvbufs++;
	UNBLOCKIO(osig);
}
#endif  /* 0 */

/*
 * sendpkt - send a packet to the specified destination
 */
void
sendpkt(dest, inter, pkt, len)
	struct sockaddr_in *dest;
	struct interface *inter;
	struct pkt *pkt;
	int len;
{
	int cc;
	extern char *ntoa();

#ifdef DEBUG
	if (debug)
		printf("sendpkt(%s, %s, %d)\n", ntoa(dest),
			ntoa(&inter->sin), len);
#endif
		
	{ struct socket dsock;
	  struct mbuf *bp;

	  /* Put the data in an mbuf.  */
	  if ( (bp = alloc_mbuf (len)) == NULLBUF)
	    {
	      inter->notsent++;
	      packets_notsent++;
	      syslog (LOG_ERR, "sendpkt: no mbuf memory");
	      return;
	    }
	  memcpy(bp->data, (char *)pkt, len);
	  bp->cnt = len;

	  dsock.address = ntohl (dest->sin_addr.s_addr);
	  dsock.port = ntohs (dest->sin_port);

	  /* send_udp() doesn't return any useful result... */
	  (void) send_udp (&inter->sock, &dsock, 0, 0, bp, 0, 0, 0);

	  inter->sent++;
	  packets_sent++;
	}
}


/*
 * input_handler- upcalled by NOS' UDP layer.
 */
void
input_handler(ifr, up, cnt)
     struct iface *ifr;
     struct udp_cb *up;
     int cnt;
{
	struct interface *interface = &inter_list[up->user];
	l_fp ts;

	handler_calls++;

	/*
	 * Grab cnt packets from the socket.  NOS makes this process much
	 * simpler than the Unix version.  I just fill in a recv_buf and
	 * upcall receive() for each packet.  If new packets arrive
	 * while I'm processing the old ones, I'll lose.  However, the
	 * correct way to fix that is to timestamp the packets at
	 * interrupt level, not to muck with lists of packets & tasks up here.
	 */

	get_systime(&ts);
	handler_pkts += cnt;

	/*
	 * Get all the packets.
	 */
	
	while (cnt--)
	  {
	    struct recvbuf rb;
	    struct socket fsocket;
	    struct sockaddr_in fsin;
	    struct mbuf *bp;
	    extern void receive();

	    rb.recv_length = recv_udp (up, &fsocket, &bp);
	    
	    if (rb.recv_length == -1) {
	      syslog(LOG_ERR, "recv_udp failed");
	      continue;
	    }
	
#ifdef DEBUG
	    if (debug > 10)
		    printf ("UDP receive from %s, length %d\n",
			    inet_ntoa (fsocket.address),
			    rb.recv_length);
#endif /* DEBUG */

	    /*
	     * Got one.  Mark how and when it got here,
	     * and upcall the receive routine.
	     */
	    /* Convert the host order fsocket to a network byte order sin. */
	    rb.recv_srcadr.sin_family = AF_INET;
	    rb.recv_srcadr.sin_port = htons (fsocket.port);
	    rb.recv_srcadr.sin_addr.s_addr = htonl (fsocket.address);

	    rb.dstadr = interface;
	    rb.recv_time = ts;
	    rb.receiver = receive;
	
	    /*
	     * Get the packet into the recv_buffer buffer.  Note that
	     * dqdata() also frees the mbuf.  If dqdata()
	     * doesn't actually copy all the bytes
	     * that we want, the upper layers will complain.  At some
	     * point, we should think about trying to do this without
	     * a copy.  Oh well.
	     */
	    rb.recv_length = dqdata (bp, &(rb.recv_buffer[0]), rb.recv_length);

	    interface->received++;
	    packets_received++;
	    
#ifdef DEBUG
	    if (debug > 20)
	      printf ("recvbuf addr: %lx\n", &rb);
#endif
	    (rb.receiver)(&rb);
	  }
}      



/*
 * findinterface - utility used by other modules to find an interface
 *		   given an address.
 */
struct interface *
findinterface(addr)
	struct sockaddr_in *addr;
{
	register int i;
	register u_long saddr;

	/*
	 * Just match the address portion.
	 */
	saddr = addr->sin_addr.s_addr;
	for (i = 0; i < ninterfaces; i++) {
		if (inter_list[i].sock.address == saddr)
			return &inter_list[i];
	}
	return (struct interface *)0;
}


/*
 * io_clr_stats - clear I/O module statistics
 */
void
io_clr_stats()
{
	packets_dropped = 0;
	packets_ignored = 0;
	packets_received = 0;
	packets_sent = 0;
	packets_notsent = 0;

	handler_calls = 0;
	handler_pkts = 0;
	io_timereset = current_time;
}


/* Simple syslog for xntp under NOS--it ignores the log level.
 */
#if defined (ANSIPROTO)
void
syslog (int level, char *fmt, ...)
{
	va_list ap;
	static char buf[512];

	va_start (ap, fmt);
	vsprintf (buf, fmt, ap);
	va_end (ap);
	log (-1, buf);
}
#else
/*VARARGS2*/
void
syslog(level,fmt,arg1,arg2,arg3,arg4,arg5)
int level;
char *fmt;
int arg1,arg2,arg3,arg4,arg5;
{
	static char buf[512];

	sprintf (buf, fmt, arg1, arg2, arg3, arg4, arg5);
	log (-1, buf);
}
#endif  /* ANSIPROTO */
