; Hardware-specific assembler routines for XNTP running under NOS
; I believe that these are specific only to the microprocessor used,
; not to the particular hardware (eg box vs. card).
	.MODEL	MEMMOD,C
	LOCALS
	%MACS
	.LALL

	; NTP-specific vars, all defined in <hardware>.c
	extrn	tickadj:dword, adjust_time:dword, ntp_tick:dword
	extrn	NClock:qword
	extrn	uticktotslo:dword:256, uticktotshi:dword:256

	; NOS clock vars
	extrn	Clock:dword, Tick:word, SavedTick:word
	extrn	clockbits:far
	
	; other needed NOS globals
	extrn	Isat:word

; ntppctick().  Called from timerproc on each clock interrupt
; (if NTP is configured into NOS).  Time is kept as an NTP
; timestamp, with the time incremented by ntp_tick for each
; clock tick.  Timerproc() saves Tick in SavedTick before clearing it,
; so we know how many clock interrupts have occurred.
;
; This code is hacked C compiler output.  See pc.c for
; the original C code.
	.CODE
	public	ntppctick
ntppctick	proc	far

	;;; Keep the old style clock going as well as NTP.
	; Clock += SavedTick;
	mov	ax,word ptr DGROUP:SavedTick
	cwd	
	add	word ptr DGROUP:Clock,ax
	adc	word ptr DGROUP:Clock+2,dx

	; for ( ; SavedTick; SavedTick--)  {
	;;; Set up the loop.  SavedTick is still in ax...
	mov	cx,ax
	;;; Punt if SavedTick is zero...
	or	cx,cx
	jnz	@70
	jmp	@87

@70:
	; if (adjust_time)  {
	mov	ax,word ptr DGROUP:adjust_time
	or	ax,word ptr DGROUP:adjust_time+2
	jne	@86
	;;; int32 adjust is always kept in dx:ax.  Clear them before
	;;; jumping to code that uses adjust.
	mov	ax,0
	mov	dx,0
	jmp	@71
@86:
	; if (adjust_time < 0)
	cmp	word ptr DGROUP:adjust_time+2,0
	jg	@72
	jl	@79
	cmp	word ptr DGROUP:adjust_time,0
	jae	@72
@79:
	; adjust = max (-tickadj, adjust_time)
	mov	dx,word ptr DGROUP:tickadj+2
	mov	ax,word ptr DGROUP:tickadj
	neg	dx
	neg	ax
	sbb	dx,0
	cmp	dx,word ptr DGROUP:adjust_time+2
	jl	@81
	jg	@73		; dx:ax already contains -tickadj, so we're done
	cmp	ax,word ptr DGROUP:adjust_time
	jae	@73		; dx:ax already contains -tickadj, so we're done
@81:
	mov	dx,word ptr DGROUP:adjust_time+2
	mov	ax,word ptr DGROUP:adjust_time
	jmp	short @73
@72:
	; adjust = min (tickadj, adjust_time);
	mov	dx,word ptr DGROUP:tickadj+2
	mov	ax,word ptr DGROUP:tickadj
	cmp	dx,word ptr DGROUP:adjust_time+2
	jg	@84
	jl	@83		; dx:ax already contains tickadj, so we're done
	cmp	ax,word ptr DGROUP:adjust_time
	jbe	@83		; dx:ax already contains tickadj, so we're done
@84:
	mov	dx,word ptr DGROUP:adjust_time+2
	mov	ax,word ptr DGROUP:adjust_time
@83:
@73:
	; adjust_time -= adjust;
	sub	word ptr DGROUP:adjust_time,ax
	sbb	word ptr DGROUP:adjust_time+2,dx
@71:
	; adjust += ntp_tick
	add	ax,word ptr DGROUP:ntp_tick
	adc	dx,word ptr DGROUP:ntp_tick+2
	; L_ADDUF (&NClock, (unsigned long)adjust);
	;;; adjust is still in dx:ax.
	;;; Just do a four word add with carry to the appropriate words of
	;;; the timestamp.  Turn off interrupts so that NClock and
	;;; SavedTick are updated atomically.  (Note that SavedTick has been
	;;; copied into cx for loop control.)
	pushf
	cli
	add	word ptr DGROUP:NClock+4,ax
	adc	word ptr DGROUP:NClock+6,dx
	adc	word ptr DGROUP:NClock,0
	adc	word ptr DGROUP:NClock+2,0
	dec	word ptr DGROUP:SavedTick
	popf

@65:
	; while SavedTick--
	dec	cx
	jz	@87
	jmp	@70

@87:
	ret	
ntppctick	endp


; getntptimeofday()---called whenever someone wants to know the current
; time in ntp format.  If on an AT, reads the hardware clock bits to
; get as accurate a time as possible.
;
; Again, this is hacked from assembler produced by the C compiler.
; See pc.c for the C code.
	public	getntptimeofday
getntptimeofday	proc	far
	push	bp
	mov	bp,sp
	sub	sp,16

	; Save si & di since I want to use them.
	push	si
	push	di

	; Make sure movsw goes in the correct direction
	cld

	; if (!Isat)  {
	cmp	word ptr DGROUP:Isat,0
	jne	@95
	; *ts = NClock;
	les	di,dword ptr [bp+6]		; load ts into es:di
	mov	si,offset DGROUP:NClock		; load &NCLOCK into ds:si
	mov	cx,4				; move 4 words
	rep movsw
	; return
	jmp	@91
	; }

@95:
	; do {
	;   hi  = Clock + Tick + SavedTick;
	;;; Assume that Tick and SavedTick are << 2**16
	mov	ax,word ptr DGROUP:Tick
	add	ax,word ptr DGROUP:SavedTick
	cwd
	add	ax,word ptr DGROUP:Clock
	adc	dx,word ptr DGROUP:Clock+2
@96:
	;;; dx:ax now contains Clock + Tick  + SavedTick
	mov	word ptr [bp-4],dx
	mov	word ptr [bp-6],ax

	; *ts = NClock;
	les	di,dword ptr [bp+6]		; (as above)
	mov	si,offset DGROUP:NClock
	mov	cx,4
	rep movsw

	; lo = clockbits();
	call	far ptr clockbits
	mov	word ptr [bp-2],ax
	; } while (hi != Clock + Tick + SavedTick);
	;;; sum Clock, Tick & SavedTick again into dx:ax
	mov	ax,word ptr DGROUP:Tick
	add	ax,word ptr DGROUP:SavedTick
	; ticks = Tick + SavedTick
	mov	word ptr [bp-8],ax
	;;; clockbits() now returns the number of unacknowledged clock
	;;; interrupts in dx.  Add that into ticks.
	add	word ptr [bp-8],dx
	cwd
	add	ax,word ptr DGROUP:Clock
	adc	dx,word ptr DGROUP:Clock+2
	;;; Compare to hi
	cmp	ax,word ptr [bp-6]
	;;; Note that Clock + Tick + SavedTick is in dx:ax, so jump to the
	;;; store, not the add.
	jne	@96
	cmp	dx,word ptr [bp-4]
	jne	@96

@101:
	; lo = -lo
	mov	ax,word ptr [bp-2]
	neg	ax
	mov	word ptr [bp-2],ax

	; hi = uticktotslo[lo & oxff];
	;;; Note that we don't ever bother storing hi in memory from here on.
	mov	bx,ax
	and	bx,255
	shl	bx,1
	shl	bx,1
	mov	dx,word ptr DGROUP:uticktotslo[bx+2]
	mov	ax,word ptr DGROUP:uticktotslo[bx]

	;hi += uticktotshi[lo >> 8]
	mov	bx,word ptr [bp-2]
	mov	cl,6
	shr	bx,cl
	and	bx,0fffch
	add	ax,word ptr DGROUP:uticktotshi[bx]
	adc	dx,word ptr DGROUP:uticktotshi[bx+2]

	; L_ADDUF (ts, hi);
	;;; Add in hi while loading ts into registers si:di:dx:ax
	les	bx,dword ptr [bp+6]
	mov	si,0
	mov	di,0
	add	ax,word ptr es:[bx+4]
	adc	dx,word ptr es:[bx+6]
	adc	di,word ptr es:[bx]
	adc	si,word ptr es:[bx+2]

	; while (ticks--)
	mov	cx,word ptr [bp-8]
	jcxz	@98

@97:
	; L_ADDUF (ts, ntp_tick)
	;;; ts is still in si:di:dx:ax
	add	ax,word ptr DGROUP:ntp_tick
	adc	dx,word ptr DGROUP:ntp_tick+2
	adc	di,0
	adc	si,0

	loop	@97

@98:
	;;; Store ts back where it came from
	mov	word ptr es:[bx+4],ax
	mov	word ptr es:[bx+6],dx
	mov	word ptr es:[bx],di
	mov	word ptr es:[bx+2],si

@91:
	pop	di
	pop	si

	mov	sp,bp
	pop	bp
	ret	
getntptimeofday	endp

	end
