/*
 * ntp_event.c - event timer support routines
 */

#include "global.h"
#include "timer.h"
#include "sockaddr.h"

#include "ntp_types.h"
#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"

/*
 * For NOS, the event timer is implemented using NOS timers.
 * The various routines are called directly by NOS when the timer
 * goes off.
 */

/*
 * Separate timer for the hourly stuff.
 */
struct timer hourly_timer;

/*
 * adjustment timer
 */
struct timer adjust_timer;

static void
handle_hourly(v)
void *v;			/* Not used... */
{
	extern void hourly_stats();
	extern u_char sys_leap;	/* from ntp_proto.c */

	/* Call the stats routine. */
	hourly_stats();

	/* If we're synched, set the DOS time to be the NTP time... */
	if (sys_leap != LEAP_NOTINSYNC)
		setdostime();

	/* Restart the timer. */
	start_timer (&hourly_timer);
}

static void
handle_adjust(v)
void *v;			/* Not used...  */
{
	extern void adj_host_clock();

	adj_host_clock();

	/* Start up the timer again. */
	start_timer (&adjust_timer);
}

/*
 * init_timer - initialize the timer data structures
 */
void
init_timer()
{
	/*
	 * Initialize...
	 */

	/* Set up the adjust_timer. */
	adjust_timer.duration = SEC2TICKS(1<<CLOCK_ADJ);
	adjust_timer.func = handle_adjust;
	start_timer (&adjust_timer);

	/* Set up the hourly timer. */
	hourly_timer.duration = SEC2TICKS(60*60);
	hourly_timer.func = handle_hourly;
	start_timer (&hourly_timer);
}

/*
 * timer_clr_stats - clear timer module stat counters
 */
void
timer_clr_stats()
{
}
