/*
 * ntpd.c - main program for the fixed point NTP daemon for NOS
 */
#include <stdio.h>

/* NOS include files */
#include "global.h"
#include "sockaddr.h"

/* XNTP include files: */
#include "ntp_types.h"
#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"

/*
 * Debugging flag
 */
int debug = 0;

/*
 * So that dontp knows whether to initialize everything or not.
 */
int initialized = 0;

/*
 * Other routines need to know that initialization is going on.
 */
int initializing = 0;


/*
 * Version declaration
 */
extern char *NTPVersion;

/*
 * ntp command for NOS.  Initialize us (if necessary), process 
 * a configuration line in the args, and return.
 */
dontp(argc, argv, p)
     int argc;
     char *argv[];
     void *p;
{
	if (!initialized)
	  {
	    syslog(LOG_INFO, NTPVersion);
	    tprintf ("%s\n", NTPVersion);

	    initializing = 1;

	    /*
	     * Call the init_ routines to initialize the data structures.
	     * Note that init_systime() may run a protocol to get a crude
	     * estimate of the time as an NTP client when running on the
	     * gizmo board.  It is important that this be run before
	     * init_subs() since the latter uses the time of day to seed
	     * the random number generator.  That is not the only
	     * dependency between these, either, be real careful about
	     * reordering.
	     */
#ifdef XNTP_AUTHENTICATE
	    init_auth();
#endif  /* XNTP_AUTHENTICATE */
	    init_util();
	    init_restrict();
	    init_loopfilter();
#ifdef XNTP_MONITOR
	    init_mon();
#endif
	    init_systime();
	    init_timer();
	    init_lib();
	    init_random();
#ifdef XNTP_MODE7
	    init_request();
#endif  /* XNTP_MODE7 */
	    init_leap();
	    init_peer();
	    init_proto();
	    init_io();

	    initializing = 0;
	  }

	/*
	 * Act on any command line arguments.
	 */
	if (argc > 1)
	  getconfig(argc, argv);

#ifdef XNTP_MODE6
	/*
	 * Report that we're up to any trappers
	 */
	if (! initialized)
	  report_event(EVNT_SYSRESTART, (struct peer *)0);
#endif  /* XNTP_MODE6 */

	initialized = 1;

	return 0;
}
